"""
# `abstract_webtools.py` Documentation

This script, `abstract_webtools.py`, is a component of the `abstract_webtools` module and is a part of the `abstract_essentials` package. It provides a set of tools and functions to interact with and parse web content.

## Contents

1. **Imports**
   - Essential modules and classes for web requests, SSL configurations, and URL parsing are imported at the beginning.

2. **Core Functions**

   - `get_status(url: str) -> int or None`:
     Fetches the HTTP status code for a given URL.

   - `clean_url(url: str) -> list`:
     Returns variations of the given URL with different protocols.

   - `get_correct_url(url: str, session: requests.Session) -> str or None`:
     Identifies the correct URL from possible variations using HTTP requests.

   - `try_request(url: str, session: requests.Session) -> requests.Response or None`:
     Attempts to make an HTTP request to a given URL.

   - `is_valid(url: str) -> bool`:
     Validates if a given URL is structurally correct.

   - `desktop_user_agents() -> list`:
     Returns a list of popular desktop user-agent strings.

   - `get_user_agent(user_agent: str) -> dict`:
     Returns a dictionary containing the user-agent header.

3. **TLSAdapter Class**

   A custom HTTPAdapter class that manages SSL options and ciphers for web requests.

   - `TLSAdapter.__init__(self, ssl_options: int)`: 
     Initializes the adapter with specific SSL options.

   - Several methods to handle cipher strings, creation of cipher strings, and initialization of the pool manager with custom SSL configurations.

4. **Advanced Web Functions**

   - `get_Source_code(url: str, user_agent: str) -> str or None`:
     Retrieves the source code of a website with a custom user-agent.

   - `parse_react_source(url: str) -> list`:
     Extracts JavaScript and JSX source code from the specified URL.

   - `get_all_website_links(url: str) -> list`:
     Lists all the internal URLs found on a specific webpage.

   - `parse_all(url: str) -> dict`:
     Parses source code to extract details about elements, attributes, and class names.

   - `extract_elements(url: str, element_type: str, attribute_name: str, class_name: str)`:
     Extracts specific portions of source code based on provided filters. The function signature seems to be cut off, so the full details aren't available.

## Usage

The functions and classes provided in this module allow users to interact with websites, from simple actions like getting the status code of a URL to more advanced functionalities such as parsing ReactJS source codes or extracting specific HTML elements from a website.

To utilize this module, simply import the required function or class and use it in your application. The functions have been designed to be intuitive and the provided docstrings give clear guidance on their usage.

Author: putkoff
Version: 1.0
"""
import requests
import ssl
from requests.adapters import HTTPAdapter
from urllib3.poolmanager import PoolManager
from abstract_utilities.time_utils import get_time_stamp,get_sleep,sleep_count_down
from abstract_utilities.json_utils import safe_json_loads
from urllib3.util import ssl_
from urllib.parse import urlparse, urljoin
from bs4 import BeautifulSoup
from urllib.parse import urlparse
import logging

logging.basicConfig(level=logging.INFO)
import logging
from abstract_utilities.time_utils import get_time_stamp, sleep_count_down

logging.basicConfig(level=logging.INFO)

class DynamicRateLimiterManager:
    def __init__(self):
        self.services = {}

    def add_service(self, service_name, low_limit, high_limit, limit_epoch, starting_tokens=None, epoch_cycle_adjustment:int=None):
        if service_name in self.services:
            logging.warning(f"Service {service_name} already exists!")
            return
        self.services[service_name] = DynamicRateLimiter(low_limit, high_limit, limit_epoch, starting_tokens, epoch_cycle_adjustment)

    def request(self, service_name):
        if service_name not in self.services:
            raise ValueError(f"Service {service_name} not found!")
        
        limiter = self.services[service_name]
        can_request = limiter.request()
        self.log_request(service_name, can_request)
        return can_request

    def log_request(self, service_name, success):
        self.services[service_name].calculate_tokens()
        logging.info(f"[{service_name}] Request {'succeeded' if success else 'denied'}. Current tokens: {self.services[service_name].tokens}")

class DynamicRateLimiter:
    def __init__(self, low_limit, high_limit, limit_epoch, starting_tokens=None, epoch_cycle_adjustment:int=None):
        self.low_limit = low_limit
        self.high_limit = high_limit
        self.limit_epoch = limit_epoch
        self.current_limit = starting_tokens or low_limit
        self.epoch_cycle_adjustment = epoch_cycle_adjustment or 0
        self.last_adjusted_time = get_time_stamp()
        self.initialize_request_status()

    def initialize_request_status(self):
        self.request_status = {
            "successful": [],
            "unsuccessful": [],
            "last_requested": get_time_stamp(),
            "epoch_initial_time": 0,
            "epoch_left": self.limit_epoch,
            "count_since_fail": 0
        }
    def request_tracker(self,success):
        if self.request_status["epoch_initial_time"]==0:
            self.request_status["epoch_initial_time"] = get_time_stamp()
        if success == True:
            self.request_status["successful"].append(get_time_stamp())
            self.request_status["count_since_fail"]=max(self.request_status["count_since_fail"]+1,1)
        else:
            self.request_status["unsuccessful"].append(get_time_stamp())
            self.request_status["last_fail"]=get_time_stamp()
            self.request_status["count_since_fail"]=min(self.request_status["count_since_fail"]-1,0)
            print(f"count since fail {self.request_status['count_since_fail']}")
        self.request_status["last_requested"]=get_time_stamp()

    def is_past_first_epoch(self):
        return (get_time_stamp() - self.request_status["epoch_initial_time"]) > self.limit_epoch

    def calculate_tokens(self):
        self.get_elapsed_time_for_current_epoch()
        elapsed_time = get_time_stamp() - self.request_status["last_requested"]
        new_tokens = int((elapsed_time / self.limit_epoch) * self.current_limit)
        self.tokens = min(self.current_limit, new_tokens)
        return self.request_status
    def adjust_rate(self):
        if self.is_past_first_epoch() and self.request_status["count_since_fail"] < -2:
            if self.request_status["count_since_fail"] % 2 == 0:
                self.limit_epoch += 1
                sleep_count_down(self.limit_epoch)

    def request(self):
        self.adjust_rate()
        self.calculate_tokens()
        if self.tokens > 0:
            return True
        return False
    def requests_in_current_epoch(self,type_request:str="successful"):
        epoch_elapsed = self.get_elapsed_time_for_current_epoch()
        epoch_left = self.limit_epoch-epoch_elapsed
        fraction_elapsed = epoch_elapsed / self.limit_epoch
        successful = []
        for each in self.request_status[type_request]:
            if (each - self.request_status["current_epoch_start"])<self.limit_epoch:
                successful.append(each)
        return len(successful)
    def get_elapsed_time_for_current_epoch(self):
        elapsed_time_in_current_epoch = 0
        current_time = get_time_stamp()
        if self.request_status["epoch_initial_time"] != None:
            elapsed_time_in_current_epoch = (current_time - self.request_status["epoch_initial_time"]) % self.limit_epoch
        self.request_status["current_epoch_start"]=current_time-elapsed_time_in_current_epoch
        self.request_status["epoch_left"]=self.limit_epoch-elapsed_time_in_current_epoch
        #print(f" epoch initial time = {self.request_status['epoch_initial_time']}\ncurrent epoch time elapsed = {elapsed_time_in_current_epoch}")
        return elapsed_time_in_current_epoch
class DynamicRateLimiterManagerSingleton:
    _instance = None
    
    @staticmethod
    def get_instance():
        if DynamicRateLimiterManagerSingleton._instance is None:
            DynamicRateLimiterManagerSingleton._instance = DynamicRateLimiterManager()
        return DynamicRateLimiterManagerSingleton._instance

def desktop_user_agents() -> list:
    """
    Returns a list of popular desktop user-agent strings for various browsers.

    Returns:
        list: A list of desktop user-agent strings.
    """
    return ['Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/89.0.4389.90 Safari/537.36',
            'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:87.0) Gecko/20100101 Firefox/87.0',
            'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.1 Safari/605.1.15',
            'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36 Edg/91.0.864.59',
            'Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:11.0) like Gecko',
            'Opera/9.80 (Windows NT 6.0) Presto/2.12.388 Version/12.14']

def get_user_agent(user_agent:str=desktop_user_agents()[0]) -> dict:
    """
    Returns the user-agent header dictionary with the specified user-agent.

    Args:
        user_agent (str, optional): The user-agent string to be used. Defaults to the first user-agent in the list.

    Returns:
        dict: A dictionary containing the 'user-agent' header.
    """
    return {"user-agent": user_agent}

class SafeRequest:
    def __init__(self, max_retries=3,last_request_time=None,request_wait_limit=1.5):
        self.session = self.initialize_session()
        self.max_retries = max_retries
        self.last_request_time = last_request_time
        self.request_wait_limit = request_wait_limit
        
    def initialize_session(self,user_agent:str= desktop_user_agents()[0]):
        s = requests.Session()
        s.cookies["cf_clearance"] = "cb4c883efc59d0e990caf7508902591f4569e7bf-1617321078-0-150"
        s.headers.update(get_user_agent(user_agent))
        # Add any other headers or cookie settings here
        adapter = TLSAdapter(ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1)
        s.mount('https://', adapter)
        return s

    @staticmethod
    def clean_url(url):
        """
        Given a URL, return a list with both HTTP and HTTPS versions.
        """
        cleaned = url.replace("http://", "").replace("https://", "")
        return [f"http://{cleaned}", f"https://{cleaned}"]

    def wait_between_requests(self):
        """
        Wait between requests based on the request_wait_limit.
        """
        if self.last_request_time:
            sleep_time = self.request_wait_limit - (get_time_stamp() - self.last_request_time)
            if sleep_time > 0:
                logging.info(f"Sleeping for {sleep_time:.2f} seconds.")
                get_sleep(sleep_time)

    def make_request(self, url, last_request_time=None, request_wait_limit=None):
        """
        Make a request and handle potential errors.
        """
        # Update the instance attributes if they are passed
        if last_request_time is not None:
            self.last_request_time = last_request_time
        if request_wait_limit is not None:
            self.request_wait_limit = request_wait_limit
            

        self.wait_between_requests()

        cleaned_urls = self.clean_url(url)
        for _ in range(self.max_retries):
            for cleaned_url in cleaned_urls:
                try:
                    response = self.session.get(cleaned_url)
                    if response.status_code == 200:
                        self.last_request_time = get_time_stamp()
                        return response
                    elif response.status_code == 429:
                        logging.warning(f"Rate limited by {cleaned_url}. Retrying...")
                        get_sleep(5)  # adjust this based on the server's rate limit reset time

                except requests.ConnectionError:
                    logging.error(f"Connection error for URL {cleaned_url}.")
                except requests.Timeout:
                    logging.error(f"Request timeout for URL {cleaned_url}.")
                except requests.RequestException as e:
                    logging.error(f"Request exception for URL {cleaned_url}: {e}")

        logging.error(f"Failed to retrieve content from {url} after {self.max_retries} retries.")
        return None


    @staticmethod
    def is_valid_url(url):
        """
        Check if the given URL is valid.
        """
        parsed = urlparse(url)
        return bool(parsed.netloc) and bool(parsed.scheme)
    def get_source_code(self, url):
        if self.is_valid_url(url):
            response = self.make_request(url)
            return response.text if response else None
        else:
            logging.error(f"Invalid URL: {url}")
            return None
## ## 
# Usage
## safe_requester = SafeRequest()
## 
## url = "example.com"  # replace with your URL
## if safe_requester.is_valid_url(url):
##     response = safe_requester.make_request(url)
##     if response:
##         print(response.text)
## else:
##     logging.error(f"Invalid URL: {url}")
# Usage 2
##    safe_requester = SafeRequest()
##    source_code = safe_requester.get_source_code('https://www.example.com')
##    if source_code:
##        print(source_code)
## ## 
class TLSAdapter(HTTPAdapter):
    """
    A custom HTTPAdapter class that sets TLS/SSL options and ciphers.

    Attributes:
        ssl_options (int): The TLS/SSL options to use when creating the SSL context.
    """
    def ssl_options(self) -> int:
        """
        Returns the SSL options to be used when creating the SSL context.

        Returns:
            int: The SSL options.
        """
        return ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1 | ssl.OP_NO_COMPRESSION

    def __init__(self, ssl_options:int=0, *args, **kwargs):
        """
        Initializes the TLSAdapter with the specified SSL options.

        Args:
            ssl_options (int, optional): The TLS/SSL options to use when creating the SSL context. Defaults to 0.
        """
        self.ssl_options = ssl_options
        super().__init__(*args, **kwargs)

    def add_string_list(self, ls: (list or str), delim: str = '', string: str = '') -> str:
        """
        Concatenates the elements of a list into a single string with the given delimiter.

        Args:
            ls (list or str): The list of elements or a comma-separated string.
            delim (str, optional): The delimiter to use when concatenating elements. Defaults to an empty string.
            string (str, optional): The initial string to append elements. Defaults to an empty string.

        Returns:
            str: The concatenated string.
        """
        if isinstance(ls, str):
            ls = list(ls.split(','))
        if isinstance(ls, list):
            string = ''
            for part in ls:
                string = string + delim + part
        return string

    def get_ciphers(self) -> str:
        """
        Returns a list of preferred TLS/SSL ciphers.

        Returns:
            list: A list of TLS/SSL ciphers.
        """
        return "ECDHE-RSA-AES256-GCM-SHA384,ECDHE-ECDSA-AES256-GCM-SHA384,ECDHE-RSA-AES256-SHA384,ECDHE-ECDSA-AES256-SHA384,ECDHE-RSA-AES256-SHA,ECDHE-ECDSA-AES256-SHA,ECDHE-RSA-AES128-GCM-SHA256,ECDHE-RSA-AES128-SHA256,ECDHE-ECDSA-AES128-GCM-SHA256,ECDHE-ECDSA-AES128-SHA256,AES256-SHA,AES128-SHA".split(',')

    def create_ciphers_string(self, ls:list=None) ->str:
        """
        Creates a colon-separated string of TLS/SSL ciphers from a list of ciphers.

        Args:
            ls (list, optional): The list of TLS/SSL ciphers to use. Defaults to None, in which case it uses the default list.

        Returns:
            str: The colon-separated string of TLS/SSL ciphers.
        """
        if ls is None:
            ls = self.get_ciphers()
        cipher_string = self.add_string_list(ls=ls, delim=':')[:-1]
        globals()['CIPHERS'] = cipher_string
        return cipher_string

    def init_poolmanager(self, *args, **kwargs) ->None:
        """
        Initializes the pool manager with the custom SSL context and ciphers.
        """
        context = ssl_.create_urllib3_context(ciphers=self.create_ciphers_string(), cert_reqs=ssl.CERT_REQUIRED, options=self.ssl_options)
        self.poolmanager = PoolManager(*args, ssl_context=context, **kwargs)
def safe_json_loads(obj):
    if isinstance(obj,(dict,list)):
        return obj
    try_obj = str(obj)
    try:
        obj = json.loads(try_obj)
    except:
        pass
    return obj
def get_token_sleep(manager, service_name, success, response):
    manager.services[service_name].request_tracker(success)
    print(f"Rate limited by {service_name}. Adjusted limit. Retrying...")
    sleep_count_down(manager.services[service_name].calculate_tokens()["token_sleep"])
    print(response)
    return True

def get_limited_request(request_url, service_name="default"):
    manager = DynamicRateLimiterManagerSingleton.get_instance()
    
    while True:
        if not manager.request(service_name):
            print(f"Rate limit for {service_name} limited by user. \nsuccessful runs = {manager.services[service_name].requests_in_current_epoch(type_request='successful')}\nWaiting for the next epoch...")
            sleep_count_down(manager.services[service_name].calculate_tokens()["epoch_left"])
            manager.services[service_name].request_status["successful"] = []
            manager.services[service_name].request_status["unsuccessful"] = []

        response = requests.get(request_url)
        if manager.services[service_name].request_status["epoch_initial_time"] is None:
            manager.services[service_name].request_status["epoch_initial_time"] = get_time_stamp()

        if response.status_code == 200:
            response_data = response.json()
            try_sleep = False

            if 'status' in response_data and isinstance(response_data, dict):
                if response_data['status'] == '0':
                    try_sleep = True

            if 'response' in response_data:
                if any(limit_message in response_data['response'].lower() for limit_message in ['max rate limit reached', "you've exceeded the rate limit"]):
                    try_sleep = get_token_sleep(manager, service_name, False, response_data)

            if not try_sleep:
                manager.services[service_name].request_tracker(True)
                return response_data
            else:
                print(response_data)

        elif response.status_code == 429:
            get_token_sleep(manager, service_name, False, response.json())

        elif response.status_code not in [200, 429]:
            print(f"Unexpected response: {response.status_code}. Message: {response.text}")
            return None

        
def strip_web(url: str) -> str:
    """
    Strip the 'http://' or 'https://' prefix from a URL, if present.

    Parameters:
    url (str): The URL string to process.

    Returns:
    str: The URL string with the prefix removed.
    """
    if url.startswith("http://"):
        url = url.replace("http://", '', 1)
    elif url.startswith("https://"):
        url = url.replace("https://", '', 1)
    return url

def get_status(url:str=None) -> int:
    """
    Gets the HTTP status code of the given URL.

    Args:
        url (str): The URL to check the status of.

    Returns:
        int: The HTTP status code of the URL, or None if the request fails.
    """
    # Get the status code of the URL
    return try_request(url=url).status_code

def clean_url(url:str=None) -> (list or None):
    """
    Cleans the given URL and returns a list of possible variations.

    Args:
        url (str): The URL to clean.

    Returns:
        list: A list of possible URL variations, including 'http://' and 'https://' prefixes.
    """
    # Clean the URL and return possible variations
    if url == None:
        return
    urls = [url]
    if url.startswith('https://'):
        urls.append('http://' + url[len('https://'):])
    elif url.startswith('http://'):
        urls.append('https://' + url[len('http://'):])
    else:
        urls.append('https://' + url)
        urls.append('http://' + url)
    return urls

def get_correct_url(url: str=None, session: type(requests.Session) = requests) -> (str or None):
    """
    Gets the correct URL from the possible variations by trying each one with an HTTP request.

    Args:
        url (str): The URL to find the correct version of.
        session (type(requests.Session), optional): The requests session to use for making HTTP requests.
            Defaults to requests.

    Returns:
        str: The correct version of the URL if found, or None if none of the variations are valid.
    """
    if url == None:
        return
    # Get the correct URL from the possible variations
    urls = clean_url(url)
    for url in urls:
        try:
            source = session.get(url)
            return url
        except requests.exceptions.RequestException as e:
            print(e)
    return None

def try_request(url:str=None, session:type(requests.Session)=requests) -> (requests.Response or None):
    """
    Tries to make an HTTP request to the given URL using the provided session.

    Args:
        url (str): The URL to make the request to.
        session (type(requests.Session), optional): The requests session to use for making HTTP requests.
            Defaults to requests.

    Returns:
        requests.Response or None: The response object if the request is successful, or None if the request fails.
    """
    if url == None:
        return
    # Try to make the HTTP request and return the response if successful
    urls = clean_url(url)
    for url in urls:
        try:
            return session.get(url)
        except requests.exceptions.RequestException as e:
            print(e)
    return None

def is_valid(url:str=None) -> bool:
    """
    Checks whether `url` is a valid URL.

    Args:
        url (str): The URL to check.

    Returns:
        bool: True if the URL is valid, False otherwise.
    """
    if url == None:
        return False
    parsed = urlparse(url)
    return bool(parsed.netloc) and bool(parsed.scheme)                      

def get_Source_code(url: str = 'https://www.example.com', user_agent:str= desktop_user_agents()[0]) -> (str or None):
    """
    Fetches the source code of the specified URL using a custom user-agent.

    Args:
        url (str, optional): The URL to fetch the source code from. Defaults to 'https://www.example.com'.
        user_agent (str, optional): The user-agent to use for the request. Defaults to the first user-agent in the list.

    Returns:
        str or None: The source code of the URL if the request is successful, or None if the request fails.
    """
    url = get_correct_url(url)
    if url is None:
        return None
    
    s = requests.Session()
    s.cookies["cf_clearance"] = "cb4c883efc59d0e990caf7508902591f4569e7bf-1617321078-0-150"
    s.headers.update(get_user_agent(user_agent))
    adapter = TLSAdapter(ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1)
    s.mount('https://', adapter)
    r = try_request(url=url, session=s)

    if r is None:
        return None
    return r.text

def parse_react_source(url:str=None) -> list:
    """
    Fetches the source code of the specified URL and extracts JavaScript and JSX source code (React components).

    Args:
        url (str): The URL to fetch the source code from.

    Returns:
        list: A list of strings containing JavaScript and JSX source code found in <script> tags.
    """
    url = get_correct_url(url)
    if url is None:
        return []
    
    data = get_Source_code(url)
    soup = BeautifulSoup(data, 'html.parser')
    script_tags = soup.find_all('script', type=lambda t: t and ('javascript' in t or 'jsx' in t))
    react_source_code = []
    for script_tag in script_tags:
        react_source_code.append(script_tag.string)
    return react_source_code

def get_all_website_links(url:str=None) -> list:
    """
    Returns all URLs that are found on the specified URL and belong to the same website.

    Args:
        url (str): The URL to search for links.

    Returns:
        list: A list of URLs that belong to the same website as the specified URL.
    """
    url = get_correct_url(url)
    if url is None:
        return []
    
    urls = [url]
    # domain name of the URL without the protocol
    domain_name = urlparse(url).netloc
    soup = BeautifulSoup(get_Source_code(url=url), "html.parser")
    for a_tag in soup.findAll("a"):
        href = a_tag.attrs.get("href")
        if href == "" or href is None:
            # href empty tag
            continue
        # join the URL if it's relative (not an absolute link)
        href = urljoin(url, href)
        parsed_href = urlparse(href)
        # remove URL GET parameters, URL fragments, etc.
        href = parsed_href.scheme + "://" + parsed_href.netloc + parsed_href.path
        if not is_valid(href):
            # not a valid URL
            continue
        if href in urls:
            # already in the set
            continue
        if domain_name not in href:
            # external link
            continue
        urls.append(href)
    return urls

def parse_all(url:str=None):
    """
    Parses the source code of the specified URL and extracts information about HTML elements, attribute values, attribute names, and class names.

    Args:
        url (str): The URL to fetch the source code from.

    Returns:
        dict: A dict containing keys: [element_types, attribute_values, attribute_names, class_names] with values as lists for keys element types, attribute values, attribute names, and class names found in the source code.
    """
    url = get_correct_url(url)
    if url is None:
        return [], [], [], []
    
    data = get_Source_code(url)
    element_types, attribute_values, attribute_names, class_names = [], [], [], []
    data = str(data).split('<')
    for k in range(1, len(data)):
        dat = data[k].split('>')[0]
        if dat[0] != '/':
            if dat.split(' ')[0] not in element_types:
                element_types.append(dat.split(' ')[0])
        dat = dat[len(dat.split(' ')[0]) + 1:].split('"')
        for c in range(1, len(dat)):
            if len(dat[c]) > 0:
                if '=' == dat[c][-1] and ' ' == dat[c][0] and dat[c] != '/':
                    if dat[c][1:] + '"' + dat[c + 1] + '"' not in attribute_values:
                        attribute_values.append(dat[c][1:] + '"' + dat[c + 1] + '"')
                    if dat[c][1:-1] not in attribute_names:
                        attribute_names.append(dat[c][1:-1])
                    if dat[c + 1] not in class_names:
                        class_names.append(dat[c + 1])
    return {"element_types":element_types, "attribute_values":attribute_values, "attribute_names":attribute_names, "class_names":class_names}
def extract_elements(url:str=None, element_type:str=None, attribute_name:str=None, class_name:str=None) -> list:
    """
    Extracts portions of the source code from the specified URL based on provided filters.

    Args:
        url (str): The URL to fetch the source code from.
        element_type (str, optional): The HTML element type to filter by. Defaults to None.
        attribute_name (str, optional): The attribute name to filter by. Defaults to None.
        class_name (str, optional): The class name to filter by. Defaults to None.

    Returns:
        list: A list of strings containing portions of the source code that match the provided filters.
    """
    url = get_correct_url(url)
    if url is None:
        return []

    data = get_Source_code(url)
    soup = BeautifulSoup(data, 'html.parser')

    elements = []

    # If no filters are provided, return the entire source code
    if not element_type and not attribute_name and not class_name:
        elements.append(str(soup))
        return elements

    # Find elements based on the filters provided
    if element_type:
        elements.extend([str(tag) for tag in soup.find_all(element_type)])

    if attribute_name:
        elements.extend([str(tag) for tag in soup.find_all(attrs={attribute_name: True})])

    if class_name:
        elements.extend([str(tag) for tag in soup.find_all(class_=class_name)])

    return elements
def get_response(response):
    if response.headers.get('content-type') == 'application/json':
        data = safe_json_loads(response.text)
        if data:
            return data.get("response", data)
    return response.text
request_manager = DynamicRateLimiterManagerSingleton.get_instance()
def get_request(request_url:str=None, request_type: str = None, request_min: int = 20, request_max: int = 30,limit_epoch: int = 60, request_start: int = None,epoch_cycle_adjustment:int=None):
    """
    Make a limited request to the ABI URL using rate-limiting.

    :param request_type: Type of the request (default is None).
    :param request_min: Minimum requests allowed in a rate-limited epoch (default is 10).
    :param request_max: Maximum requests allowed in a rate-limited epoch (default is 30).
    :param limit_epoch: Length of the rate-limited epoch in seconds (default is 60).
    :param request_start: Start of the rate-limited epoch (default is None).
    :param json_data: JSON data for the request (default is None).
    :return: Limited response from the ABI URL.
    """
    manager = DynamicRateLimiterManagerSingleton.get_instance()
    if request_type not in manager.services:
        request_manager.add_service(request_type, request_min, request_max, limit_epoch, request_start,epoch_cycle_adjustment=epoch_cycle_adjustment)
    return get_limited_request(request_url=request_url, service_name=request_type)
def get_response(request=None):
    """
    Parse the JSON response and return the ABI.

    :return: Parsed ABI response.
    """
    if "result" in request:
        return safe_json_loads(request["result"])
    return safe_json_loads(request)

while True:
    abi = get_request(request_url="https://api.coingecko.com/api/v3/simple/price?ids=usdc&vs_currencies=usd",request_min=15,request_max=30,limit_epoch=60,request_start=30,epoch_cycle_adjustment=5)
    print(abi)



"""
# `abstract_webtools.py` Documentation

This script, `abstract_webtools.py`, is a component of the `abstract_webtools` module and is a part of the `abstract_essentials` package. It provides a set of tools and functions to interact with and parse web content.

## Contents

1. **Imports**
   - Essential modules and classes for web requests, SSL configurations, and URL parsing are imported at the beginning.

2. **Core Functions**

   - `get_status(url: str) -> int or None`:
     Fetches the HTTP status code for a given URL.

   - `clean_url(url: str) -> list`:
     Returns variations of the given URL with different protocols.

   - `get_correct_url(url: str, session: requests.Session) -> str or None`:
     Identifies the correct URL from possible variations using HTTP requests.

   - `try_request(url: str, session: requests.Session) -> requests.Response or None`:
     Attempts to make an HTTP request to a given URL.

   - `is_valid(url: str) -> bool`:
     Validates if a given URL is structurally correct.

   - `desktop_user_agents() -> list`:
     Returns a list of popular desktop user-agent strings.

   - `get_user_agent(user_agent: str) -> dict`:
     Returns a dictionary containing the user-agent header.

3. **TLSAdapter Class**

   A custom HTTPAdapter class that manages SSL options and ciphers for web requests.

   - `TLSAdapter.__init__(self, ssl_options: int)`: 
     Initializes the adapter with specific SSL options.

   - Several methods to handle cipher strings, creation of cipher strings, and initialization of the pool manager with custom SSL configurations.

4. **Advanced Web Functions**

   - `get_Source_code(url: str, user_agent: str) -> str or None`:
     Retrieves the source code of a website with a custom user-agent.

   - `parse_react_source(url: str) -> list`:
     Extracts JavaScript and JSX source code from the specified URL.

   - `get_all_website_links(url: str) -> list`:
     Lists all the internal URLs found on a specific webpage.

   - `parse_all(url: str) -> dict`:
     Parses source code to extract details about elements, attributes, and class names.

   - `extract_elements(url: str, element_type: str, attribute_name: str, class_name: str)`:
     Extracts specific portions of source code based on provided filters. The function signature seems to be cut off, so the full details aren't available.

## Usage

The functions and classes provided in this module allow users to interact with websites, from simple actions like getting the status code of a URL to more advanced functionalities such as parsing ReactJS source codes or extracting specific HTML elements from a website.

To utilize this module, simply import the required function or class and use it in your application. The functions have been designed to be intuitive and the provided docstrings give clear guidance on their usage.

Author: putkoff
Version: 1.0
"""
# -*- coding: UTF-8 -*-
import requests
import os
# Google Chrome Driver
from selenium import webdriver
import yt_dlp
import ssl
import requests
from requests.adapters import HTTPAdapter
from typing import Optional, List
from requests.packages.urllib3.poolmanager import PoolManager
from requests.packages.urllib3.util import ssl_
from urllib.parse import urlparse
import logging
import xml.etree.ElementTree as ET
from abstract_utilities.time_utils import get_time_stamp,get_sleep,sleep_count_down
from abstract_utilities.string_clean import eatInner,eatAll
logging.basicConfig(level=logging.INFO)
class DynamicRateLimiterManager:
    def __init__(self):
        # Key: Service Name, Value: DynamicRateLimiter instance
        self.services = {}
    
    def add_service(self, service_name, low_limit, high_limit, limit_epoch, starting_tokens=None):
        if service_name in self.services:
            print(f"Service {service_name} already exists!")
            return
        self.services[service_name] = DynamicRateLimiter(low_limit, high_limit, limit_epoch, starting_tokens)
    
    def request(self, service_name):
        if service_name not in self.services:
            raise ValueError(f"Service {service_name} not found!")
        
        limiter = self.services[service_name]
        can_request = limiter.request()
        
        # Log the outcome of the request attempt
        self.log_request(service_name, can_request)
        
        return can_request
    
    def log_request(self, service_name, success):
        # Placeholder logging method, replace with actual logging implementation
        print(f"[{service_name}] Request {'succeeded' if success else 'denied'}. Current tokens: {self.services[service_name].get_current_tokens()}")
class DynamicRateLimiter:
    def __init__(self, low_limit, high_limit, limit_epoch, starting_tokens=None,epoch_cycle_adjustment:int=None):
        self.low_limit = low_limit
        self.high_limit = high_limit
        self.limit_epoch = limit_epoch  # in seconds
        self.request_status_json = {"succesful":[],"unsuccesful":[],"last_requested":get_time_stamp(),"first_requested":get_time_stamp(),"epoch_left":self.limit_epoch,"last_fail":get_time_stamp(),"count_since_fail":0}
        self.current_limit = starting_tokens or low_limit  # Default to high_limit if starting_tokens isn't provided
        self.epoch_cycle_adjustment = epoch_cycle_adjustment
        # Additional attributes for tracking adjustment logic
        self.last_adjusted_time = get_time_stamp()
        self.successful_epochs_since_last_adjustment = 0
        self.request_count_in_current_epoch = 0

    def _refill_tokens(self):
        time_since_last_request = get_time_stamp() - self.request_status_json["last_requested"]
        new_tokens = (time_since_last_request / self.limit_epoch) * self.current_limit
        self.tokens = min(self.current_limit, self.get_current_tokens())
    def request_tracker(self,success):
        if success:
            self.request_status_json["succesful"].append(get_time_stamp())
        else:
            self.request_status_json["unsuccesful"].append(get_time_stamp())
            self.request_status_json["last_fail"]=get_time_stamp()
            self.request_status_json["count_since_fail"]=0
            self.adjust_limit()
        self.request_status_json["last_requested"]=get_time_stamp()
    def calculate_tokens(self):
        successful = []
        for each in self.request_status_json["succesful"]:
            if (get_time_stamp() - each)<self.limit_epoch:
                successful.append(each)
        self.request_status_json["succesful"]=successful
        unsuccessful = []
        for each in self.request_status_json["unsuccesful"]:
            if (get_time_stamp() - each)<self.limit_epoch:
                unsuccessful.append(each)
        self.request_status_json["unsuccesful"]=unsuccessful
        if len(successful)==0 and len(unsuccessful)==0:
            pass
        elif len(successful)!=0 and len(unsuccessful)==0:
            self.request_status_json["first_requested"] = successful[0]
        elif len(successful)==0 and len(unsuccessful)!=0:
            self.request_status_json["first_requested"] = unsuccessful[0]
        else:
            self.request_status_json["first_requested"] = min(unsuccessful[0],successful[0])
        self.request_status_json["epoch_left"]=self.limit_epoch-(self.request_status_json["last_requested"]-self.request_status_json["first_requested"])
        
        return self.request_status_json
    def get_current_tokens(self):
        self.request_status_json = self.calculate_tokens()
        total_requests = len(self.request_status_json["succesful"])+len(self.request_status_json["unsuccesful"])
        return max(0,self.current_limit-total_requests)
    def get_sleep(self):
        self.request_status_json = self.calculate_tokens()
        self.request_status_json["current_sleep"]=self.request_status_json["epoch_left"]/max(1,self.get_current_tokens())
        return self.request_status_json
    def request(self):
        self._refill_tokens()
        if self.tokens > 0:
            return True  # The request can be made
        else:
            if self.tokens == 0:
                self.request_status_json["count_since_fail"]+=1
                if self.epoch_cycle_adjustment != None:
                    if self.request_status_json["count_since_fail"] >=self.epoch_cycle_adjustment:
                        self.current_limit=min(self.current_limit+1,self.high_limit)
            return False  # The request cannot be made
    def _adjust_limit(self):
        current_time = get_time_stamp()
        if current_time - self.last_adjusted_time >= self.limit_epoch:
            if len(self.clear_epoch()["succesful"]) >= self.tokens:
                # We hit the rate limit this epoch, decrease our limit
                self.tokens = max(1, self.tokens - 1)
            else:
                self.successful_epochs_since_last_adjustment += 1
                if self.successful_epochs_since_last_adjustment >= 5:
                    # We've had 5 successful epochs, increase our limit
                    self.current_limit = min(self.high_limit, self.tokens + 1)
                    self.successful_epochs_since_last_adjustment = 0
            
            # Reset our counters for the new epoch
            self.last_adjusted_time = current_time
            self.request_count_in_current_epoch = 0
    def adjust_limit(self):
        # Set the tokens to succesful requests_made - 1
        self.tokens = len(self.calculate_tokens()["succesful"])

        # Adjust the high_limit
        self.current_limit = self.tokens

        # Log the adjustment
        print(f"Adjusted tokens to: {self.tokens} and high_limit to: {self.current_limit}")
class DynamicRateLimiterManagerSingleton:
    _instance = None
    @staticmethod
    def get_instance():
        if DynamicRateLimiterManagerSingleton._instance is None:
            DynamicRateLimiterManagerSingleton._instance = DynamicRateLimiterManager()
        return DynamicRateLimiterManagerSingleton._instance

class CipherManager:
    @staticmethod
    def  get_default_ciphers()-> list:
        return [
            "ECDHE-RSA-AES256-GCM-SHA384", "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-SHA384", "ECDHE-ECDSA-AES256-SHA384",
            "ECDHE-RSA-AES256-SHA", "ECDHE-ECDSA-AES256-SHA",
            "ECDHE-RSA-AES128-GCM-SHA256", "ECDHE-RSA-AES128-SHA256",
            "ECDHE-ECDSA-AES128-GCM-SHA256", "ECDHE-ECDSA-AES128-SHA256",
            "AES256-SHA", "AES128-SHA"
        ]

    def __init__(self,cipher_list=None):
        if cipher_list == None:
            cipher_list=self.get_default_ciphers()
        self.cipher_list = cipher_list
        self.create_list()
        self.ciphers_string = self.add_string_list()
    def add_string_list(self):
        if len(self.cipher_list)==0:
            return ''
        return','.join(self.cipher_list)
    def create_list(self):
        if self.cipher_list == None:
            self.cipher_list= []
        elif isinstance(self.cipher_list, str):
            self.cipher_list=self.cipher_list.split(',')
        if isinstance(self.cipher_list, str):
            self.cipher_list=[self.cipher_list]
class CipherManagerSingleton:
    _instance = None
    @staticmethod
    def get_instance(cipher_list=None):
        if CipherManagerSingleton._instance is None:
            CipherManagerSingleton._instance = CipherManager(cipher_list=cipher_list)
        elif CipherManagerSingleton._instance.cipher_list != cipher_list:
            CipherManagerSingleton._instance = CipherManager(cipher_list=cipher_list)
        return CipherManagerSingleton._instance
class SSLManager:
    @staticmethod
    def get_default_certification():
        return ssl.CERT_REQUIRED

    @staticmethod
    def get_default_tls_options():
        return ["OP_NO_TLSv1", "OP_NO_TLSv1_1", "OP_NO_COMPRESSION"]

    @staticmethod
    def get_all_tls_options() -> int:
        """
        Returns the SSL options to be used when creating the SSL context.
            [
         ssl.OP_SINGLE_ECDH_USE,
         ssl.OP_SINGLE_DH_USE,
         ssl.OP_NO_TLSv1_3,
         ssl.OP_NO_TLSv1_2,
         ssl.OP_NO_TLSv1_1,
         ssl.OP_NO_TLSv1,
         ssl.OP_NO_TICKET,
         ssl.OP_NO_RENEGOTIATION,
         ssl.OP_NO_QUERY_MTU,
         ssl.OP_NO_COMPRESSION,
         ssl.OP_CIPHER_SERVER_PREFERENCE,
         ssl.OP_ALLOW_NO_DHE_KEX,
         ssl.OP_ALL
         ]
         The `ssl` module in the Python standard library provides several constants that you can use to set various SSL options. Here are the available options as of Python 3.9:

        1. `ssl.OP_ALL`:
           - Enables a collection of various bug workaround options.

        2. `ssl.OP_ALLOW_NO_DHE_KEX`:
           - Allow a non-(EC)DHE handshake on a server socket if no suitable security level can be reached.

        3. `ssl.OP_CIPHER_SERVER_PREFERENCE`:
           - Uses the server's cipher ordering preference rather than the client's.

        4. `ssl.OP_NO_COMPRESSION`:
           - Prevents using SSL/TLS compression to avoid CRIME attacks.

        5. `ssl.OP_NO_QUERY_MTU`:
           - Disables automatic querying of kernel for MTU.

        6. `ssl.OP_NO_RENEGOTIATION`:
           - Disallows all renegotiation.

        7. `ssl.OP_NO_TICKET`:
           - Disables use of RFC 5077 session tickets.

        8. `ssl.OP_NO_TLSv1`:
           - Prevents the use of TLSv1.

        9. `ssl.OP_NO_TLSv1_1`:
           - Prevents the use of TLSv1.1.

        10. `ssl.OP_NO_TLSv1_2`:
           - Prevents the use of TLSv1.2.

        11. `ssl.OP_NO_TLSv1_3`:
           - Prevents the use of TLSv1.3.

        12. `ssl.OP_SINGLE_DH_USE`:
           - Always create a new key when using temporary/ephemeral DH parameters. This option provides forward secrecy.

        13. `ssl.OP_SINGLE_ECDH_USE`:
           - Always create a new key when using temporary/ephemeral ECDH parameters. This option provides forward secrecy.

        These constants can be combined using the bitwise OR (`|`) operator to set multiple options. For example, to prevent the use of TLSv1 and TLSv1.1, you would use:
        Please note that the availability of some options might vary depending on the version of OpenSSL that Python's `ssl` module is linked against and the version of Python itself. You can always check the Python documentation specific to your version to get the most accurate and updated list.

        Returns:
            int: The SSL options.

        """
        return [
            "OP_SINGLE_ECDH_USE",
            "OP_SINGLE_DH_USE",
            "OP_NO_TLSv1_3",
            "OP_NO_TLSv1_2",
            "OP_NO_TLSv1_1",
            "OP_NO_TLSv1",
            "OP_NO_TICKET",
            "OP_NO_RENEGOTIATION",
            "OP_NO_QUERY_MTU",
            "OP_NO_COMPRESSION",
            "OP_CIPHER_SERVER_PREFERENCE",
            "OP_ALLOW_NO_DHE_KEX",
            "OP_ALL"
            ]

    @staticmethod
    def get_context(ciphers=None, options=None, cert_reqs=None):
        
        return ssl_.create_urllib3_context(ciphers=ciphers, cert_reqs=cert_reqs, options=options)

    def __init__(self, ciphers=None, ssl_options_list=None, certification=None):
        self.ssl_options_list = ssl_options_list
        self.create_list()
        self.ssl_options_values = self.get_options_values()
        self.ssl_options = self.combine_ssl_options()
        self.certification = certification or self.get_default_certification()
        self.cipher_manager = CipherManagerSingleton().get_instance(cipher_list=ciphers)
        self.ssl_context = self.get_context(ciphers=self.cipher_manager.ciphers_string, options=self.ssl_options, cert_reqs=self.certification)
    def create_list(self):
        if self.ssl_options_list == None:
            self.ssl_options_list= []
        elif isinstance(self.ssl_options_list, str):
            self.ssl_options_list=self.ssl_options_list.split(',')
        if isinstance(self.ssl_options_list, str):
            self.ssl_options_list=[self.ssl_options_list]
    def get_options_values(self):
        return [getattr(ssl, option_name) for option_name in self.ssl_options_list]
    def combine_ssl_options(self):
        combined_options = 0
        for option in self.ssl_options_values:
            combined_options |= option
        return combined_options
class SSLManagerSingleton:
    _instance = None
    @staticmethod
    def get_instance(ciphers=None, ssl_options_list=None, certification=None):
        if SSLManagerSingleton._instance is None:
            SSLManagerSingleton._instance = SSLManager(ciphers=ciphers, ssl_options_list=ssl_options_list, certification=certification)
        elif SSLManagerSingleton._instance.cipher_manager.ciphers_string != ciphers or SSLManagerSingleton._instance.ssl_options_list !=ssl_options_list or SSLManagerSingleton._instance.certification !=certification:
            SSLManagerSingleton._instance = SSLManager(ciphers=ciphers, ssl_options_list=ssl_options_list, certification=certification)
        return SSLManagerSingleton._instance
class TLSAdapter(HTTPAdapter):
    def __init__(self, ciphers: Optional[List[str]] = None, certification: Optional[str] = None, ssl_options: Optional[List[str]] = None):
        super().__init__()
        self.ciphers = ciphers
        self.certification = certification
        self.ssl_options = ssl_options
        
        self.cipher_manager = CipherManagerSingleton.get_instance(cipher_list=self.ciphers)
        self.ssl_manager = SSLManagerSingleton.get_instance(
            ciphers=self.cipher_manager.ciphers_string, 
            ssl_options_list=ssl_options, 
            certification=certification
        )
        self.ssl_context = self.ssl_manager.ssl_context

    def init_poolmanager(self, *args, **kwargs):
        return super().init_poolmanager(*args, **kwargs)


class TLSAdapterSingleton:
    _instance = None
    @staticmethod
    def get_instance(ciphers=None, certification=None, ssl_options=None):
        if TLSAdapterSingleton._instance is None:
            TLSAdapterSingleton._instance = TLSAdapter(ciphers=ciphers, certification=certification, ssl_options=ssl_options)
        elif TLSAdapterSingleton._instance.ciphers != ciphers or SSLManagerSingleton._instance.certification !=certification or SSLManagerSingleton._instance.ssl_options_list !=ssl_options:
            TLSAdapterSingleton._instance = TLSAdapter(ciphers=ciphers, certification=certification, ssl_options=ssl_options)
        return TLSAdapterSingleton._instance
class UserAgentManager:
    @staticmethod
    def desktop_user_agents() -> list:
        """
        Returns a list of popular desktop user-agent strings for various browsers.

        Returns:
            list: A list of desktop user-agent strings.
        """
        return ['Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/89.0.4389.90 Safari/537.36',
                'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:87.0) Gecko/20100101 Firefox/87.0',
                'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.1 Safari/605.1.15',
                'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36 Edg/91.0.864.59',
                'Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:11.0) like Gecko',
                'Opera/9.80 (Windows NT 6.0) Presto/2.12.388 Version/12.14']
    @staticmethod
    def get_user_agent(user_agent:str=desktop_user_agents()[0]) -> dict:
        """
        Returns the user-agent header dictionary with the specified user-agent.

        Args:
            user_agent (str, optional): The user-agent string to be used. Defaults to the first user-agent in the list.

        Returns:
            dict: A dictionary containing the 'user-agent' header.
        """
        return {"user-agent": user_agent}
    def __init__(self,user_agent=desktop_user_agents()[0]):
        self.user_agent = user_agent
        self.user_agent_header=self.get_user_agent(user_agent=user_agent)
class UserAgentManagerSingleton:
    _instance = None
    @staticmethod
    def get_instance(user_agent=UserAgentManager.desktop_user_agents()[0]):
        if UserAgentManagerSingleton._instance is None:
            UserAgentManagerSingleton._instance = UserAgentManager(user_agent=user_agent)
        elif UserAgentManagerSingleton._instance.user_agent != user_agent:
            UserAgentManagerSingleton._instance = UserAgentManager(user_agent=user_agent)
        return UserAgentManagerSingleton._instance
class SafeRequest:
    def __init__(self,
                 url=None,
                 headers:dict=UserAgentManagerSingleton().get_instance().user_agent_header,
                 max_retries=3,
                 last_request_time=None,
                 request_wait_limit=1.5,
                 ):
        self.url =url 
        if isinstance(headers,str):
            headers = UserAgentManagerSingleton().get_instance(user_agent=headers).user_agent_header
        self.headers = headers
        self.max_retries=max_retries
        
        self.request_wait_limit = request_wait_limit
        
        self.url_manager = URLManagerSingleton().get_instance(url=url)
        self.session = self.initialize_session()
        
        self.last_request_time = last_request_time
        
        self.response = self.make_request()
        
        self.status_code = None if self.response ==  None else self.response.status_code
        
        self.source_code = '' if self.response == None else self.response.text
    def initialize_session(self):
        s = requests.Session()
        s.cookies["cf_clearance"] = "cb4c883efc59d0e990caf7508902591f4569e7bf-1617321078-0-150"
        s.headers.update(self.headers)
        # Add any other headers or cookie settings here
        adapter = TLSAdapterSingleton().get_instance()
        s.mount('https://', adapter)
        return s

    @staticmethod
    def clean_url(url):
        """
        Given a URL, return a list with potential URL versions.
        """
        cleaned = url.replace("http://", "").replace("https://", "")
        no_subdomain = cleaned.replace("www.", "", 1)

        urls = [
            f"https://{cleaned}",
            f"http://{cleaned}",
        ]

        # Add variants without 'www' if it was present
        if cleaned != no_subdomain:
            urls.extend([
                f"https://{no_subdomain}",
                f"http://{no_subdomain}",
            ])

        return urls

    def wait_between_requests(self):
        """
        Wait between requests based on the request_wait_limit.
        """
        if self.last_request_time:
            sleep_time = self.request_wait_limit - (get_time_stamp() - self.last_request_time)
            if sleep_time > 0:
                logging.info(f"Sleeping for {sleep_time:.2f} seconds.")
                get_sleep(sleep_time)

    def make_request(self, last_request_time=None, request_wait_limit=None):
        """
        Make a request and handle potential errors.
        """
        # Update the instance attributes if they are passed
        if last_request_time is not None:
            self.last_request_time = last_request_time
        if request_wait_limit is not None:
            self.request_wait_limit = request_wait_limit
            
        
        self.wait_between_requests()
        
        for _ in range(self.max_retries):
            try:
                response = self.session.get(url=self.url_manager.correct_url, timeout=10)  # 10 seconds timeout
                if response.status_code == 200:
                    self.last_request_time = get_time_stamp()
                    return response
                elif response.status_code == 429:
                    logging.warning(f"Rate limited by {self.url_manager.correct_url}. Retrying...")
                    get_sleep(5)  # adjust this based on the server's rate limit reset time
            except requests.Timeout as e:
                logging.error(f"Request to {cleaned_url} timed out: {e}")
            except requests.ConnectionError:
                logging.error(f"Connection error for URL {self.url_manager.correct_url}.")
            except requests.Timeout:
                logging.error(f"Request timeout for URL {self.url_manager.correct_url}.")
            except requests.RequestException as e:
                logging.error(f"Request exception for URL {self.url_manager.correct_url}: {e}")

        logging.error(f"Failed to retrieve content from {self.url_manager.correct_url} after {self.max_retries} retries.")
        return None
    @staticmethod
    def is_valid_url(url):
        """
        Check if the given URL is valid.
        """
        parsed = urlparse(url)
        return bool(parsed.netloc) and bool(parsed.scheme)
    def get_source_code(self, url=None,response=None):
        if self.response:
            return self.response.text if self.response else None
        else:
            logging.error(f"Invalid URL: {url}")
            return None
        self.clean_url(self.url_manager.correct_url)
class SafeRequestSingleton:
    _instance = None
    @staticmethod
    
    def get_instance(url=None,headers:dict=UserAgentManager().user_agent_header,max_retries=3,last_request_time=None,request_wait_limit=1.5):
        if SafeRequestSingleton._instance is None:
            SafeRequestSingleton._instance = SafeRequest(url=url,headers=headers,max_retries=max_retries,last_request_time=last_request_time,request_wait_limit=request_wait_limit)
        elif SafeRequestSingleton._instance.url != url or SafeRequestSingleton._instance.headers != headers or SafeRequestSingleton._instance.max_retries != max_retries or SafeRequestSingleton._instance.request_wait_limit != request_wait_limit:
            SafeRequestSingleton._instance = SafeRequest(url=url,headers=headers,max_retries=max_retries,last_request_time=last_request_time,request_wait_limit=request_wait_limit)
        return SafeRequestSingleton._instance
## ## 
# Usage
## safe_requester = SafeRequest()
## 
## url = "example.com"  # replace with your URL
## if safe_requester.is_valid_url(url):
##     response = safe_requester.make_request(url)
##     if response:
##         print(response.text)
## else:
##     logging.error(f"Invalid URL: {url}")
# Usage 2
##    safe_requester = SafeRequest()
##    source_code = safe_requester.get_source_code('https://www.example.com')
##    if source_code:
##        print(source_code)
## ##
class URLManager:
    def __init__(self,url=None,session=requests):
        self.url = url
        self.session = session
        self.striped_url = None if url ==  None else self.strip_web()
        self.clean_urls = None if url ==  None else self.clean_url()
        self.correct_url = None if url ==  None else self.get_correct_url()
    def strip_web(self) -> str:
        """
        Strip the 'http://' or 'https://' prefix from a URL, if present.

        Parameters:
        url (str): The URL string to process.

        Returns:
        str: The URL string with the prefix removed.
        """
        url = self.url
        if self.url.startswith("http://"):
            url = self.url.replace("http://", '', 1)
        elif self.url.startswith("https://"):
            url = self.url.replace("https://", '', 1)
        return url
    def clean_url(self) -> (list or None):
        """
        Cleans the given URL and returns a list of possible variations.

        Args:
            url (str): The URL to clean.

        Returns:
            list: A list of possible URL variations, including 'http://' and 'https://' prefixes.
        """
        # Clean the URL and return possible variations
        urls = [self.url]
        if self.url.startswith('https://'):
            urls.append('http://' + self.url[len('https://'):])
        elif self.url.startswith('http://'):
            urls.append('https://' + self.url[len('http://'):])
        else:
            urls.append('https://' + self.url)
            urls.append('http://' + self.url)
        return urls

    def get_correct_url(self) -> (str or None):
        """
        Gets the correct URL from the possible variations by trying each one with an HTTP request.

        Args:
            url (str): The URL to find the correct version of.
            session (type(requests.Session), optional): The requests session to use for making HTTP requests.
                Defaults to requests.

        Returns:
            str: The correct version of the URL if found, or None if none of the variations are valid.
        """
        # Get the correct URL from the possible variations
        for url in self.clean_urls:
            try:
                source = self.session.get(url)
                return url
            except requests.exceptions.RequestException as e:
                print(e)
        return None
class URLManagerSingleton:
    _instance = None
    @staticmethod
    def get_instance(url=None,session=requests):
        if URLManagerSingleton._instance is None:
            URLManagerSingleton._instance = URLManager(url=url,session=session)
        elif URLManagerSingleton._instance.session != session or URLManagerSingleton._instance.url != url:
            URLManagerSingleton._instance = URLManager(url=url,session=session)
        return URLManagerSingleton._instance

class VideoDownloader:
    
    def __init__(self, url,title=None,download_directory=os.getcwd(),user_agent=None,video_extention='mp4'):
        self.url = url
        self.video_extention=video_extention
        self.header = UserAgentManagerSingleton().get_instance(user_agent=user_agent).user_agent_header
        self.base_name = os.path.basename(self.url)
        self.file_name,self.ext = os.path.splitext(self.base_name)
        self.download_directory=download_directory
        self.title = url.split('/')[3] if title == None else title
        self.video_urls = []
        self.fetch_video_urls()
        self.download_videos()
    def fetch_video_urls(self):
        driver = webdriver.Chrome()
        driver.get(self.url)
        self.page_source = driver.page_source
        for each in self.page_source.split('<source ')[1:]:
            # NOTE: Make sure to import the `eatAll` function and use it here.
            self.video_urls.append(eatInner(each.split('.{self.video_extention}'.replace('..','.'))[0].split('http')[-1],['h','t','t','p','s',':','//','/','s','=',' ','\n','\t',''])+'.mp4')
    def download_videos(self):
        for video_url in self.video_urls:
            ydl_opts = {}
            with yt_dlp.YoutubeDL(ydl_opts) as ydl:
                info = ydl.extract_info(self.url)
                self.base_name = os.path.basename(info['url'])
                self.file_name,self.ext = os.path.splitext(self.base_name)
                video_content =SafeRequestSingleton().get_instance(url=info['url']).response
                print("Start downloading")
                content_length = int(video_content.headers['content-length'])
                print(f'Size: {content_length / 1024 / 1024:.2f}MB')
                down_size = 0
                with open(f'{os.path.join(self.download_directory,self.base_name)}', "wb") as video:
                    for chunk in video_content.iter_content(chunk_size=1024 * 1024):
                        if chunk:
                            video.write(chunk)
                            down_size += len(chunk)
                            print(f'Progress: {down_size / content_length:.2%}', end='\r')
class VideoDownloaderSingleton():
    _instance = None
    @staticmethod
    def get_instance(url,title=None,video_extention='mp4',download_directory=os.getcwd(),user_agent=None):
        if VideoDownloaderSingleton._instance is None:
            VideoDownloaderSingleton._instance = VideoDownloader(url=url,parse_type=parse_type)
        elif VideoDownloaderSingleton._instance.title != title or video_extention != VideoDownloaderSingleton._instance.video_extention or url != VideoDownloaderSingleton._instance.url or download_directory != VideoDownloaderSingleton._instance.download_directory or user_agent != VideoDownloaderSingleton._instance.user_agent:
            VideoDownloaderSingleton._instance = VideoDownloader(url=url,parse_type=parse_type)
        return VideoDownloaderSingleton._instance
class SoupManager:
    def __init__(self,url=None,source_code=None,parse_type="html.parser"):
        self.url = url
        self.source_code = source_code
        self.parse_type = parse_type
        if self.source_code ==None:
            if self.url != None:
                self.url_manager = URLManagerSingleton().get_instance(url=url)
                self.request_manager = SafeRequestSingleton().get_instance(url=self.url_manager .correct_url)
                self.source_code = self.request_manager.source_code
            else:
                return 
        self.comments = []
        self.pure_source = self.un_comment_source_code()
        self.attributes = self.get_all_attributes()
        self.class_names=[]
        self.class_values=[]
        self.attribute_tracker_js={}
        self.get_all_class_names()
        self.soup = BeautifulSoup(self.source_code, parse_type)
        self.meta_tags = self.get_meta_tags()
    def find_all(self,element,soup=None):
        soup = self.soup if soup == None else soup
        return soup.find_all(element)
    def get_class(self,class_name,soup=None):
        soup = self.soup if soup == None else soup
        return soup.get(class_name)
    def get_meta_tags(self):
        new_soup = self.find_all("meta")
        meta_tags={}
        for meta_component in new_soup:
            meta_property = self.get_class(soup = meta_component,class_name="property")
            if meta_property != None:
                if meta_property not in meta_tags:
                    meta_tags[meta_property] = []
                content = self.get_class(soup = meta_component,class_name="content")
                if content != None and content not in  meta_tags[meta_property]:
                    meta_tags[meta_property].append(content)
        return meta_tags
    def get_rid_of_sections(self,source_code,comment_delimeters=['!--','-->']):
        begin_comment=source_code.split(comment_delimeters[0])
        source = begin_comment[0]
        for section in begin_comment[1:]:
            if comment_delimeters[1] in section:
                end_section = section.split(comment_delimeters[1])[0]+comment_delimeters[1]
                self.comments.append(comment_delimeters[0]+end_section)
                source+=section[len(end_section):]
        return source
    def un_comment_source_code(self):
        comment_delimeters = ['!--','-->']
        comments = []
        pure_source_code=self.get_rid_of_sections(self.source_code,comment_delimeters=comment_delimeters)
        comment_delimeters = ['/*','*/']
        pure_source_code=self.get_rid_of_sections(pure_source_code,comment_delimeters=comment_delimeters)
        return pure_source_code
    def get_all_attributes(self):
        attributes=[]
        sources = self.pure_source.replace('<','\n<').split('\n<')
        for source in sources:
            attribute = source.split(' ')[0].split('>')[0]
            attribute=eatAll(attribute,['',' ','\n','\t'])
            if len(attribute)>0:
                if '/' == attribute[0]:
                    attribute = attribute[1:]
                if attribute not in attributes:
                    attributes.append(attribute)
        return attributes
    def determine_if_quotes_open(self,quotes,char):
        quote_tracker = quotes[char]
        if quote_tracker[0] !=  quote_tracker[1]:
           if (quote_tracker[0]-quote_tracker[1])<0:
               quote_tracker[0]+=1
           else:
                quote_tracker[1]+=1
        elif quote_tracker[0] ==  quote_tracker[1]:
            quote_tracker[0]+=1
        for each in list(quotes.keys()):
            if each != char:
                other_quote_tracker = quotes[each]
        if other_quote_tracker[0]  !=   other_quote_tracker[1] or quote_tracker[0] !=  quote_tracker[1]:
            return quotes,True
        if other_quote_tracker[0]  ==   other_quote_tracker[1] and quote_tracker[0] ==  quote_tracker[1]:
            return quotes,False
    def eat_till_quote(self, component_str):
            while component_str and component_str[0] not in ['"', "'"]:
                component_str = component_str[1:]
            if component_str:
                initial_quote = component_str[0]
                component_str = component_str[1:]
                k = 0
                while k < len(component_str) and component_str[k] != initial_quote:
                    k += 1
                return component_str[:k], component_str[k+1:]
            return None, None
    def find_class_name(self, string):
        while string and string[0] in [' ', '\t', '\n']:
            string = string[1:]
        if '=' in string:
            class_name, string = string.split('=', 1)
            class_name = class_name.strip()
            if class_name.startswith('"') or class_name.startswith("'"):
                class_name = class_name[1:]
            return class_name.strip(), string.strip()
        return None, None
    def get_all_class_names(self):
        for attribute in self.attributes:
            attrib_sources = self.pure_source.split(f'<{attribute} ')
            for attrib_source in attrib_sources[1:]:
                class_items = attrib_source.split('>')[0]
                while class_items != None:
                    class_name, class_items = self.find_class_name(class_items)
                    class_value, class_items = self.eat_till_quote(class_items)
                    if class_name and class_value:
                        if class_name not in self.attribute_tracker_js:
                            self.attribute_tracker_js[class_name] = {"class_values":[],"attributions":[]}
                        if class_name not in self.class_names:
                            self.class_names.append(class_name)
                        if class_value not in self.class_values:
                           self.class_values.append(class_value)
                        if class_value not in self.attribute_tracker_js[class_name]["class_values"]:
                            self.attribute_tracker_js[class_name]["class_values"].append(class_value)
                        if attribute not in self.attribute_tracker_js[class_name]["attributions"]:
                            self.attribute_tracker_js[class_name]["attributions"].append(attribute)
class SoupManagerSingleton():
    _instance = None
    @staticmethod
    def get_instance(url,parse_type="html.parser"):
        if SoupManagerSingleton._instance is None:
            SoupManagerSingleton._instance = SoupManager(url=url,parse_type=parse_type)
        elif parse_type != SoupManagerSingleton._instance.parse_type or url != SoupManagerSingleton._instance.url :
            SoupManagerSingleton._instance = SoupManager(url=url,parse_type=parse_type)
        return SoupManagerSingleton._instance
def get_limited_request(request_url=str,service_name="default"):
    manager = DynamicRateLimiterManagerSingleton.get_instance()  # Get the singleton instance
    unwanted_response=True
    # Check with the rate limiter if we can make a request
    while True:
        if not manager.request(service_name):
            print("Rate limit reached for coin_gecko. Waiting for the next epoch...")
            sleep_count_down(manager.services[service_name].get_sleep()["current_sleep"])  # Wait for the limit_epoch duration
        # Make the actual request
        response = requests.get(request_url)
        
        # If you get a rate-limit error (usually 429 status code but can vary), adjust the rate limiter
        if response.status_code == 429:
            print(response.json())
            manager.services[service_name].request_tracker(False)
            print("Rate limited by coin_gecko. Adjusted limit. Retrying...")
            if len(manager.services[service_name].calculate_tokens()["succesful"])<2:
                sleep_count_down(manager.services[service_name].limit_epoch)  # Wait for the limit_epoch duration
            else:
                manager.services[service_name].current_limit-=1
                sleep_count_down(manager.services[service_name].limit_epoch/len(manager.services[service_name].calculate_tokens()["succesful"]))  # Wait for the limit_epoch duration
        # Return the data if the request was successful
        if response.status_code == 200:
            manager.services[service_name].request_tracker(True)
            return response.json()
        elif response.status_code not in [200,429]:
            print(f"Unexpected response: {response.status_code}. Message: {response.text}")
            return None



def get_status(url:str=None) -> int:
    """
    Gets the HTTP status code of the given URL.

    Args:
        url (str): The URL to check the status of.

    Returns:
        int: The HTTP status code of the URL, or None if the request fails.
    """
    # Get the status code of the URL
    return try_request(url=url).status_code


def try_request(url:str=None, session:type(requests.Session)=requests) -> (requests.Response or None):
    """
    Tries to make an HTTP request to the given URL using the provided session.

    Args:
        url (str): The URL to make the request to.
        session (type(requests.Session), optional): The requests session to use for making HTTP requests.
            Defaults to requests.

    Returns:
        requests.Response or None: The response object if the request is successful, or None if the request fails.
    """
    if url == None:
        return
    # Try to make the HTTP request and return the response if successful
    urls = clean_url(url)
    for url in urls:
        try:
            return session.get(url)
        except requests.exceptions.RequestException as e:
            print(e)
    return None

def is_valid(url:str=None) -> bool:
    """
    Checks whether `url` is a valid URL.

    Args:
        url (str): The URL to check.

    Returns:
        bool: True if the URL is valid, False otherwise.
    """
    if url == None:
        return False
    parsed = urlparse(url)
    return bool(parsed.netloc) and bool(parsed.scheme)                      

def get_Source_code(url: str ='https://www.example.com' , user_agent:str= UserAgentManager().desktop_user_agents()[0]) -> (str or None):
    """
    Fetches the source code of the specified URL using a custom user-agent.

    Args:
        url (str, optional): The URL to fetch the source code from. Defaults to 'https://www.example.com'.
        user_agent (str, optional): The user-agent to use for the request. Defaults to the first user-agent in the list.

    Returns:
        str or None: The source code of the URL if the request is successful, or None if the request fails.
    """
    url_manager = URLManagerSingleton().get_source_code(url=url)
    if url_manager.url is None:
        return []
    s = requests.Session()
    s.cookies["cf_clearance"] = "cb4c883efc59d0e990caf7508902591f4569e7bf-1617321078-0-150"
    s.headers.update(get_user_agent(user_agent))
    adapter = TLSAdapter()
    s.mount('https://', adapter)
    r = try_request(url=url_manager.correct_url, session=s)

    if r is None:
        return None
    return r.text

def parse_react_source(url:str=None) -> list:
    """
    Fetches the source code of the specified URL and extracts JavaScript and JSX source code (React components).

    Args:
        url (str): The URL to fetch the source code from.

    Returns:
        list: A list of strings containing JavaScript and JSX source code found in <script> tags.
    """
    url_manager = URLManagerSingleton().get_source_code(url=url)
    if url_manager.url is None:
        return []
    request_manager = SafeRequestSingleton().get_instance(url=url_manager.url)
    soup = BeautifulSoup(request_manager.source_code, "html.parser")
    script_tags = soup.find_all('script', type=lambda t: t and ('javascript' in t or 'jsx' in t))
    react_source_code = []
    for script_tag in script_tags:
        react_source_code.append(script_tag.string)
    return react_source_code

def get_all_website_links(url:str=None) -> list:
    """
    Returns all URLs that are found on the specified URL and belong to the same website.

    Args:
        url (str): The URL to search for links.

    Returns:
        list: A list of URLs that belong to the same website as the specified URL.
    """
    request_manager = SafeRequestSingleton().get_instance(url=url)
    if request_manager.url is None:
        return []
    
    
    urls = [request_manager.url]
    # domain name of the URL without the protocol
    domain_name = urlparse(request_manager.url).netloc
    soup = BeautifulSoup(request_manager.source_code, "html.parser")
    for a_tag in soup.findAll("a"):
        href = a_tag.attrs.get("href")
        if href == "" or href is None:
            # href empty tag
            continue
        # join the URL if it's relative (not an absolute link)
        href = urljoin(url, href)
        parsed_href = urlparse(href)
        # remove URL GET parameters, URL fragments, etc.
        href = parsed_href.scheme + "://" + parsed_href.netloc + parsed_href.path
        if not is_valid(href):
            # not a valid URL
            continue
        if href in urls:
            # already in the set
            continue
        if domain_name not in href:
            # external link
            continue
        urls.append(href)
    return urls

def parse_all(url:str=None):
    """
    Parses the source code of the specified URL and extracts information about HTML elements, attribute values, attribute names, and class names.

    Args:
        url (str): The URL to fetch the source code from.

    Returns:
        dict: A dict containing keys: [element_types, attribute_values, attribute_names, class_names] with values as lists for keys element types, attribute values, attribute names, and class names found in the source code.
    """
    request_manager = SafeRequest().get_source_code(url=url)
    if request_manager.url is None:
        return [], [], [], []
    
    
    soup = BeautifulSoup(request_manager.source_code, 'html.parser')
    element_types, attribute_values, attribute_names, class_names = [], [], [], []
    data = str(data).split('<')
    for k in range(1, len(data)):
        dat = data[k].split('>')[0]
        if dat[0] != '/':
            if dat.split(' ')[0] not in element_types:
                element_types.append(dat.split(' ')[0])
        dat = dat[len(dat.split(' ')[0]) + 1:].split('"')
        for c in range(1, len(dat)):
            if len(dat[c]) > 0:
                if '=' == dat[c][-1] and ' ' == dat[c][0] and dat[c] != '/':
                    if dat[c][1:] + '"' + dat[c + 1] + '"' not in attribute_values:
                        attribute_values.append(dat[c][1:] + '"' + dat[c + 1] + '"')
                    if dat[c][1:-1] not in attribute_names:
                        attribute_names.append(dat[c][1:-1])
                    if dat[c + 1] not in class_names:
                        class_names.append(dat[c + 1])
    return {"element_types":element_types, "attribute_values":attribute_values, "attribute_names":attribute_names, "class_names":class_names}
def extract_elements(url:str=None, element_type:str=None, attribute_name:str=None, class_name:str=None) -> list:
    """
    Extracts portions of the source code from the specified URL based on provided filters.

    Args:
        url (str): The URL to fetch the source code from.
        element_type (str, optional): The HTML element type to filter by. Defaults to None.
        attribute_name (str, optional): The attribute name to filter by. Defaults to None.
        class_name (str, optional): The class name to filter by. Defaults to None.

    Returns:
        list: A list of strings containing portions of the source code that match the provided filters.
    """
    request_manager = SafeRequest().get_source_code(url=url)
    if request_manager.url is None:
        return [], [], [], []
    soup = BeautifulSoup(request_manager.source_code, 'html.parser')
    elements = []

    # If no filters are provided, return the entire source code
    if not element_type and not attribute_name and not class_name:
        elements.append(str(soup))
        return elements

    # Find elements based on the filters provided
    if element_type:
        elements.extend([str(tag) for tag in soup.find_all(element_type)])

    if attribute_name:
        elements.extend([str(tag) for tag in soup.find_all(attrs={attribute_name: True})])

    if class_name:
        elements.extend([str(tag) for tag in soup.find_all(class_=class_name)])

    return elements
def get_response(response):
    if response.headers.get('content-type') == 'application/json':
        data = safe_json_loads(response.text)
        if data:
            return data.get("response", data)
    return response.text



import requests
import ssl
from bs4 import BeautifulSoup
from requests.adapters import HTTPAdapter
from urllib3.poolmanager import PoolManager
from urllib3.util import ssl_
from abstract_utilities import *
import PySimpleGUI as sg
import inspect
import re
from .abstract_webtools import URLManagerSingleton,CipherManagerSingleton,SafeRequestSingleton,UserAgentManagerSingleton
from abstract_utilities.class_utils import call_functions,process_args,get_fun,mk_fun
window = None
def get_request_manager(url=None,):
  return SafeRequestSingleton().get_instance(url=url)
def get_url_manager(url=None):
  return URLManagerSingleton().get_instance(url=url)
def get_soup_manager(url=None,selected_option='html.parser',delim=""):
  return SoupManagerSingleton().get_instance(url=url,selected_option=selected_option,delim=delim)
def get_user_agent_manager(user_agent=None):
  return UserAgentManagerSingleton().get_instance(user_agent=user_agent)
def get_cipher_manager(cipher_list=None):
  return CipherManagerSingleton().get_instance(cipher_list=cipher_list)
def get_url():
  if window is None:
    return 'example.com'
  event, values = window.read()
  return values['-URL-']
class SoupManager:
  @staticmethod
  def get_parser_choices():
    return ['html.parser', 'lxml', 'html5lib']
  def __init__(self,url=None,selected_option='html.parser',delim="a"):
    self.url=url
    self.selected_option = selected_option
    self.url_manager = get_url_manager(url=self.url)
    self.request_manager = SafeRequestSingleton().get_instance(url=self.url_manager.correct_url)
    self.soup = BeautifulSoup(self.request_manager.source_code, self.selected_option)
    self.find_all = self.soup.find_all(delim)
class SoupManagerSingleton:
    _instance = None
    @staticmethod
    def get_instance(url=None,selected_option='html.parser',delim=""):
        if SoupManagerSingleton._instance is None:
            SoupManagerSingleton._instance = SoupManager(url=url,selected_option=selected_option,delim=delim)
        elif SoupManagerSingleton._instance.url != url:
            SoupManagerSingleton._instance = SoupManager(url=url,selected_option=selected_option,delim=delim)
        return SoupManagerSingleton._instance
def get_gui_fun(name:str='',args:dict={}):
  import PySimpleGUI
  return get_fun({"instance":PySimpleGUI,"name":name,"args":args})
def expandable(size:tuple=(None,None)):
    return {"size": size,"resizable": True,"scrollable": True,"auto_size_text": True,"expand_x":True,"expand_y": True}
def change_glob(var:any,val:any):
    globals()[var]=val
    return val
def get_parser_choices():
    bs4_module = inspect.getmodule(BeautifulSoup)
    docstring = bs4_module.__builtins__
    start_index = docstring.find("Supported parsers are:")
    end_index = docstring.find(")", start_index)
    choices_text = docstring[start_index:end_index]
    choices = [choice.strip() for choice in choices_text.split(",")]
    return choices
def add_string_list(ls:(list or str),delim:str='',string:str=''):
    for k in range(0,len(list(ls))):
        string=string+str(list(ls)[k])+delim
    return string
def get_ciphers():
    return "ECDHE-RSA-AES256-GCM-SHA384,ECDHE-ECDSA-AES256-GCM-SHA384,ECDHE-RSA-AES256-SHA384,ECDHE-ECDSA-AES256-SHA384,ECDHE-RSA-AES256-SHA,ECDHE-ECDSA-AES256-SHA,ECDHE-RSA-AES128-GCM-SHA256,ECDHE-RSA-AES128-SHA256,ECDHE-ECDSA-AES128-GCM-SHA256,ECDHE-ECDSA-AES128-SHA256,AES256-SHA,AES128-SHA".split(',')
def create_ciphers_string(ls:list=get_ciphers()):
    cipher_string = add_string_list(ls=ls,delim=':')[:-1]
    print(cipher_string)
    globals()['CIPHERS']=cipher_string
    return cipher_string
create_ciphers_string()

def get_browsers():
    return 'Chrome,Firefox,Safari,Microsoft Edge,Internet Explorer,Opera'.split(',')
def get_user_agents():
    return ['Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/89.0.4389.90 Safari/537.36','Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:87.0) Gecko/20100101 Firefox/87.0','Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.1 Safari/605.1.15','Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36 Edg/91.0.864.59','Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:11.0) like Gecko','Opera/9.80 (Windows NT 6.0) Presto/2.12.388 Version/12.14']
def create_user_agent(user_agent:str=get_user_agents()[0]):
    return {"user-agent": user_agent}
def get_operating_systems():
  return ['Windows NT 10.0','Macintosh; Intel Mac OS X 10_15_7','Linux','Android','iOS']
def ssl_options():
    return ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1 | ssl.OP_NO_COMPRESSION
def get_user_agent():
    return {"user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/89.0.4389.90 Safari/537.36"}
class TLSAdapter(HTTPAdapter):
    def __init__(self, ssl_options=0, *args, **kwargs):
        self.ssl_options = ssl_options
        super().__init__(*args, **kwargs)

    def init_poolmanager(self, *args, **kwargs):
        context = ssl_.create_urllib3_context(ciphers=CIPHERS, cert_reqs=ssl.CERT_REQUIRED, options=self.ssl_options)
        self.poolmanager = PoolManager(*args, ssl_context=context, **kwargs)
def create_columns(ls,i,k):
    if float(i)%float(k)==float(0.00) and i != 0:
        lsN = list(ls[:-k])
        lsN.append(list(ls[-k:]))
        ls = lsN
    return ls
def get_cypher_checks(ciphers:list=get_ciphers()):
    ls=[[[sg.Text('CIPHERS: ')],sg.Multiline(CIPHERS,key='-CIPHERS_OUTPUT-', size=(80, 5), disabled=False)]]
    for k in range(0,len(ciphers)):
        ls.append(sg.Checkbox(ciphers[k],key=ciphers[k],default=True,enable_events=True))
        ls = create_columns(ls,k,5)
    return ls
def format_url(url):
    # Ensure the URL starts with 'http://' or 'https://'
    if not url.startswith(('http://', 'https://')):
        url = 'https://' + url
    return url
def try_request(url: str, session:type(requests.Session)=requests):
    try:
        return session.get(url)
    except requests.exceptions.RequestException as e:
        print(e)
        return None


def get_parsed_html(url:str='https://www.example.com', header:str=create_user_agent()):
    s = requests.Session()
    s.cookies["cf_clearance"] = "cb4c883efc59d0e990caf7508902591f4569e7bf-1617321078-0-150"
    s.headers.update({"user-agent": get_user_agents()[0]})
    adapter = TLSAdapter(ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1)
    s.mount('https://', adapter)
    r = try_request(url=url, session=s)
    data = r.text
    if r is False:
        return None
    data = r.text
    return data
def parse_all(data):
    ls_type,ls_tag,ls_class,ls_desc =[],[],[],[]
    data = str(data).split('<')
    for k in range(1,len(data)):
        dat = data[k].split('>')[0]
        if dat[0] != '/':
            if dat.split(' ')[0] not in ls_type:
                ls_type.append(dat.split(' ')[0])
        dat = dat[len(dat.split(' ')[0])+1:].split('"')
        for c in range(1,len(dat)):
            if len(dat[c])>0:
                if '=' == dat[c][-1] and ' ' == dat[c][0] and dat[c] != '/':
                    if dat[c][1:]+'"'+dat[c+1]+'"' not in ls_desc:
                        ls_desc.append(dat[c][1:]+'"'+dat[c+1]+'"')
                    if dat[c][1:-1] not in ls_tag:
                        ls_tag.append(dat[c][1:-1])
                    if dat[c+1] not in ls_class:
                        ls_class.append(dat[c+1])
    return ls_type,ls_desc,ls_tag,ls_class

def parse_react_source(data):
    soup = BeautifulSoup(data, 'html.parser')
    script_tags = soup.find_all('script', type=lambda t: t and ('javascript' in t or 'jsx' in t))
    react_source_code = []
    for script_tag in script_tags:
        react_source_code.append(script_tag.string)
    return react_source_code
def all_soup(data,tag,typ,clas,inp):
    return getattr(get_soup_manager(url=get_url()).soup,tag,typ)
def find_all_soup(string:str):
    return get_soup_manager(url=get_url(),delim=string).find_all
def get_bs4_options():
    bs4_options = [
        'BeautifulSoup',
        'Tag',
        'NavigableString',
        'Comment',
        'ResultSet',
        'SoupStrainer',
        'CData'
    ]
    descriptions = [
        'The main BeautifulSoup class used for parsing HTML.',
        'Represents an HTML tag.',
        'Represents a string within an HTML document.',
        'Represents an HTML comment.',
        'Represents a collection of tags found during a search.',
        'Allows parsing only a specific subset of the HTML document.',
        'Represents a CDATA section within an XML document.'
    ]
    return list(zip(bs4_options, descriptions))
def get_multi_line(args):
    return get_gui_fun(name='Multiline',args={**args,**expandable()})
def get_gpt_layout():
    # Add a dropdown for selecting BeautifulSoup parsing capabilities
    parser_choices = ['html.parser', 'lxml', 'html5lib']
    sg.theme('LightGrey1')
    layout = [[sg.Text('URL:', size=(8, 1)), sg.Input('www.example.com', key='-URL-',enable_events=True),sg.Text('status:'),sg.Text(try_request(url='https://www.example.com').status_code,key="-STATUS_CODE-"),
               sg.Text('warning: https://www.example.com is valid'),sg.Text('',key="-URL_WARNING-"),sg.Button('correct url',key='-CORRECT_URL-',visible=True)],
        [sg.Checkbox('Custom User-Agent', default=False, key='-CUSTOMUA-', enable_events=True)],
        [sg.Text('User-Agent:', size=(8, 1)), sg.Combo(get_user_agents(), default_value=get_user_agents()[0], key='-USERAGENT-', disabled=False)],
        [get_cypher_checks()],
        [sg.Button('Grab URL'), sg.Button('Action')],
        [get_multi_line({"key":"-SOURCECODE-"})],
        [sg.Text('Parsing Capabilities:', size=(15, 1)), sg.DropDown(parser_choices, default_value='html.parser', key='-PARSER-',enable_events=True)],
        [get_multi_line({"key":"-SOUP_OUTPUT-"})],
        [sg.Text('find soup:'),[[sg.Checkbox('',default=True,key='-CHECK_TAG-',enable_events=True),sg.Combo([], size=(15, 1),key='-SOUP_TAG-',enable_events=True)],
                                [sg.Checkbox('',default=False,key='-CHECK_ELEMENT-',enable_events=True),sg.Combo([], size=(15, 1),key='-SOUP_ELEMENT-',enable_events=True)],
                                [sg.Checkbox('',default=False,key='-CHECK_TYPE-',enable_events=True),sg.Combo([], size=(15, 1),key='-SOUP_TYPE-',enable_events=True)],
                                [sg.Checkbox('',default=False,key='-CHECK_CLASS-',enable_events=True),sg.Combo([], size=(15, 1),key='-SOUP_CLASS-',enable_events=True)],
                                sg.Input(key='-SOUP_INPUT-'), sg.Button('get soup'),sg.Button('all soup')]],
              
        [get_multi_line({"key":"-FIND_ALL_OUTPUT-"})]]
    return layout
def display_soup(window,values,soup_manager):
    event, values = window.read()
    if soup_manager.soup != None:
        #window['-REACT_OUTPUT-'].update(value=parse_react_source(source_code))
        window['-SOUP_OUTPUT-'].update(value=soup_manager.soup )
        ls_type,ls_desc,ls_tag,ls_class = parse_all(soup_manager.soup)
        if len(ls_type)>0:
          window['-SOUP_TAG-'].update(values=ls_type)
        if len(ls_desc)>0:
            window['-SOUP_ELEMENT-'].update(values=ls_desc,value=ls_desc[0])
        if len(ls_tag) >0:
            window['-SOUP_TYPE-'].update(values=ls_tag,value=ls_tag[0])
        if len(ls_class) >0:
           window['-SOUP_CLASS-'].update(values=ls_class,value=ls_class[0])
    window['-FIND_ALL_OUTPUT-'].update(value=soup_manager.find_all)
def get_selected_cipher_list():
  ls = []
  event, values = window.read()
  for k in range(len(get_ciphers())):
      if values[get_ciphers()[k]] == True:
          ls.append(get_ciphers()[k])
  return ls
def process_url(window,values):
    url = values['-URL-']
    delim= values['-SOUP_INPUT-']
    selected_option = values['-PARSER-']
    user_agent=values['-USERAGENT-']
    soup_input =values['-SOUP_TAG-']
    cipher_list = get_selected_cipher_list()
    try:
      url_manager = get_url_manager(url)
      cipher_manager = get_cipher_manager(cipher_list=cipher_list)
      request_manager = get_request_manager(url=url_manager.correct_url)
      soup_manager = get_soup_manager(url=url,selected_option=selected_option,delim=delim)
      user_agent_manager = get_user_agent_manager(user_agent=user_agent)
      window['-STATUS_CODE-'].update(value=request_manager.status_code)
      window['-SOURCECODE-'].update(value=request_manager.source_code)
      window['-SOUP_OUTPUT-'].update(value=soup_manager.soup)
      window['-CIPHERS_OUTPUT-'].update(value=cipher_manager.ciphers_string)
    except:
      print(url)
  

    return url_manager,request_manager,soup_manager,user_agent_manager,cipher_manager
def url_grabber_while(window):
    while True:

      
        event, values = window.read()
        if event == sg.WINDOW_CLOSED:
            break
        if 'SOUP_' in event:
            soup_manager = get_soup_manager(url=url_manager.correct_url,selected_option=selected_option,delim=values['-SOUP_INPUT-'])
            name = event.split("SOUP_")[-1][:-1]
            check_name = f'-CHECK_{name}-'
            window[check_name].update(value=True)
            for each in values.keys():
              if 'CHECK_' in each and each != check_name:
                window[each].update(value=False)
            window[check_name].update(value=find_all_soup(values[event]))
        cipher_list=get_selected_cipher_list()
        url_manager,request_manager,soup_manager,user_agent_manager,cipher_manager=process_url(window,values)
        if event=='-CORRECT_URL-':
          if url_manager.correct_url:
            window['-URL-'].update(value=url_manager.correct_url)
        if event == 'all soup':
          selected_option = values['-PARSER-']
          soup_manager = get_soup_manager(url=url_manager.correct_url,selected_option=selected_option,delim=values['-SOUP_INPUT-'])
          display_soup(window,values,soup_manager)
          
        if event == 'all soup':
            delim= values['-SOUP_INPUT-']
            window['-FIND_ALL_OUTPUT-'].update(value=soup_manager.find_all)
        
        elif event == 'get soup':
            window['-FIND_ALL_OUTPUT-'].update(value=all_soup(values['-SOUP_OUTPUT-'],values['-SOUP_TAG-'],values['-SOUP_TYPE-'],values['-SOUP_CLASS-'],values['-SOUP_INPUT-']))
        if event == '-CUSTOMUA-':
            window['-SOURCECODE-'].update(disabled=values['-CUSTOMUA-'])
            if not values['-CUSTOMUA-']:
                
                window['-USERAGENT-'].update(value=user_agent_manager.user_agent_header)
                window['-USERAGENT-'].update(disabled=True)
            else:
                window['-USERAGENT-'].update(disabled=False)
        if event == 'Grab URL':
            if url_manager.correct_url:
                # Perform actions with the formatted URL
                process_url(window,values)
            else:
                # Invalid URL format, display an error message
                sg.popup('Invalid URL format. Please enter a valid URL.')
        if event == 'Action':
            selected_option = values['-PARSER-']
            if selected_option == 'BeautifulSoup':
                result = soup
            elif selected_option == 'Tag':
                result = soup.find('tag_name')
            elif selected_option == 'NavigableString':
                result = soup.find('tag_name').string
            elif selected_option == 'Comment':
                result = soup.find(text=lambda text: isinstance(text, Comment))
            elif selected_option == 'ResultSet':
                result = soup.find_all('tag_name')
            elif selected_option == 'SoupStrainer':
                result = BeautifulSoup(source_code, 'html.parser', parse_only=SoupStrainer('tag_name'))
            elif selected_option == 'CData':
                result = soup.find(text=lambda text: isinstance(text, CData))
            else:
                result = None
            window['-SOUP_OUTPUT-'].update(value=str(result))

    
def url_grabber_component():
    globals()['curr_check']='TAG-'
    layout = get_gpt_layout()
    globals()['window'] = get_gui_fun(name='Window',args={'title':'URL Grabber', 'layout':layout,**expandable()})
    url_grabber_while(window)

    
url_grabber_component()
